/*  signal.c
 *
 *  APT signal processing functions of xwxapt application
 */

/*
 *  xwxapt: An application to decode APT signals from
 *  weather satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "xwxapt.h"
#include "interface.h"
#include "support.h"
#include <math.h>

extern GtkWidget
  *xwxapt_main_window, /* xwxapt top window  */
  *gbl_level_pbar,     /* Audio level prog. bar */
  *gbl_sync_pbar,      /* Sync level prog. bar  */
  *gbl_status_table,   /* Status widgets' table */
  *gbl_sync_icon;      /* Sync status indicator */

/* Pointer to pixbuf pixels */
extern guchar *gbl_pixel_buf;

extern int
  gbl_sat_type, /* Weather satellite type number   */
  gbl_sync_ref, /* Required position of sync train */
  gbl_sync_err, /* Deviation from sync reference   */
  gbl_line_idx; /* Index to audio samples buffer   */

/* List of required sync position references */
extern int gbl_sync_refs[];

/* Audio/DSP samples buffer */
extern short gbl_line_buffer[];

/*------------------------------------------------------------------------*/

/*  Detect_NOAA_Sync()
 *
 *  Detects the Channel A sync of a NOAA APT signal.
 *  This is 7 cycles at 1040 Hz on the 2.4 KHz sub-carrier.
 */

  int
Detect_NOAA_Sync( int search_limit )
{
  /* FIFO Buffer for ch A sync detection */
  static int sync_fifo[NOAA_SYNC_FIFO_SIZE];

  short
  	sample_a,    /* First of the pair of carrier samples  */
	sample_b;    /* Second of the pair of carrier samples */

  int
	sync_idx,    /* Counter used in detecting sync train position  */
	carrier_val, /* Instantaneous value of 2.4 Khz sub-carrier     */
	sync_val,    /* Value produced by sync detection algorithm     */
	sync_max,    /* Max value produced by sync detection algorithm */
	sync_ave,    /* Average value of carrier samples in sync fifo  */
	sync_pos,    /* Position in sample buffer where sync o/p peaks */
	peak_idx,    /* Index used in searching for peak in sync value */
	fifo_idx,    /* Index for sync detection fifo */
	sync_cnt,	 /* Counter of sync pulses processed */
	idx;         /* Index for various loops etc.  */

  /* Move through buffer up to search_limit, looking   */
  /* for sync detector peak. On exit, the buffer index */
  /* should be at the end of 'A' sync pulse sequence.  */

  sync_max = fifo_idx = 0;
  sync_pos = gbl_line_idx;
  search_limit--;

  for( peak_idx = 0; peak_idx < search_limit; peak_idx++ )
  {
	/* Read first of the pair of carrier samples */
	sample_a = gbl_line_buffer[gbl_line_idx++];

	/* Read second of the pair of carrier samples */
	sample_b = gbl_line_buffer[gbl_line_idx];

	/* This is the instantaneous magnitude of the 2.4 Khz  */
	/* subcarrier, calculated as the scalar magnitude of   */
	/* two consecutive audio samples at fi deg phase angle */
	carrier_val = Carrier_Amplitude( sample_a, sample_b );

	/* Store carrier amplitude samples in fifo in cyclic manner */
	if( ++fifo_idx >= NOAA_SYNC_FIFO_SIZE )
	  fifo_idx = 0;
	sync_fifo[fifo_idx] = carrier_val;

	/* Evaluate the average of the samples in fifo, to */
	/* be used as a zero reference in summation below. */
	sync_val = 0;
	for( idx = 0; idx < NOAA_SYNC_FIFO_SIZE; idx++ )
	  sync_val += sync_fifo[idx];
	sync_ave = sync_val/NOAA_SYNC_FIFO_SIZE;

	/* Add and subtract NOAA_SYNC_SAMPLES samples to the  */
	/* accumulator, for each half cycle of the sync train */
	sync_val = 0;
	sync_idx = fifo_idx;
	for( sync_cnt = 0; sync_cnt < NOAA_SYNC_PULSES; sync_cnt++ )
	{
	  /* Add carrier samples into accumulator */
	  for( idx = 0; idx < NOAA_SYNC_SAMPLES; idx++ )
	  {
		sync_val += sync_fifo[sync_idx] - sync_ave;
		if( ++sync_idx >= NOAA_SYNC_FIFO_SIZE )
		  sync_idx = 0;
	  }

	  /* Subtract carrier samples from accumulator */
	  for( idx = 0; idx < NOAA_SYNC_SAMPLES; idx++ )
	  {
		sync_val -= sync_fifo[sync_idx] - sync_ave;
		if( ++sync_idx >= NOAA_SYNC_FIFO_SIZE )
		  sync_idx = 0;
	  }

	} /* for( sync_cnt = 0; sync_cnt < NOAA_SYNC_PULSES; sync_cnt++ ) */

	/* Summate the tail end of the sync train */
	for( sync_cnt = 0; sync_cnt < NOAA_SYNC_TAIL; sync_cnt++ )
	{
	  sync_val -= sync_fifo[sync_idx] - sync_ave;
	  if( ++sync_idx >= NOAA_SYNC_FIFO_SIZE )
		sync_idx = 0;
	}

	/* Record position in buffer where sync detector peaks */
	if( sync_val > sync_max )
	{
	  sync_max = sync_val;
	  sync_pos = gbl_line_idx;
	}

  } /* End of for( peak_idx = 0; peak_idx < search_limit; peak_idx++ ) */

  /* Decoding from DSP */
  if( isFlagClear(ACTION_PROCESS_FILE) )
  {
	/* Adjust sync detector progress bar */
	gdouble pbar_frac = (gdouble)sync_max/(gdouble)NOAA_SYNC_UPPER_THRESH;
	if( pbar_frac > 1.0 ) pbar_frac = 1.0;
	gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR(gbl_sync_pbar), pbar_frac );

	/* Initialize skew label */
	gtk_label_set_text( GTK_LABEL(lookup_widget(
			xwxapt_main_window, "err_label")), "?" );

	/* Setup buffer index to max sync position */
	/* if sync maximum is within the threshold */
	if( (sync_max > NOAA_SYNC_LOWER_THRESH) &&
		(sync_max < NOAA_SYNC_UPPER_THRESH) )
	{
	  char buff[12];

	  /* Display sync error */
	  gbl_line_idx = sync_pos;
	  gbl_sync_err = sync_pos - gbl_sync_ref;
	  snprintf( buff, 12, "%d", gbl_sync_err );
	  gtk_label_set_text( GTK_LABEL(lookup_widget(
			  xwxapt_main_window, "err_label")), buff );

	  /* If sync position is in range, show 'Lock' */
	  if( (abs( gbl_sync_err ) < SYNC_DETECT) &&
		  isFlagClear(ACTION_SYNC_BUFFER) )
	  {
		if( isFlagClear(ICON_SYNC_LOCK) )
		{
		  ClearFlag( ICON_FLAGS_ALL );
		  SetFlag( ICON_SYNC_LOCK );
		  Set_Sync_Icon( "gtk-yes" );
		}
	  }
	  else /* Show 'Search' */
	  {
		if( isFlagClear(ICON_SYNC_SEARCH) )
		{
		  ClearFlag( ICON_FLAGS_ALL );
		  SetFlag( ICON_SYNC_SEARCH );
		  Set_Sync_Icon( "gtk-dialog-question" );
		}

	  } /* if( abs(gbl_line_idx - gbl_sync_ref) < 3 ) */
	}
	else /* If sync level is out of range, show 'Unlock' */
	{
	  gbl_line_idx = 0;

	  if( isFlagClear(ICON_SYNC_UNLOCK) )
	  {
		ClearFlag( ICON_FLAGS_ALL );
		SetFlag( ICON_SYNC_UNLOCK );
		Set_Sync_Icon( "gtk-no" );
	  }

	} /* if( (sync_max > NOAA_SYNC_LOWER_THRESH) && */

  } /* if( isFlagClear(ACTION_PROCESS_FILE) ) */
  else
  {
	/* Setup buffer index to max sync position */
	/* if sync maximum is within the threshold */
	if( (sync_max > NOAA_SYNC_LOWER_THRESH) &&
		(sync_max < NOAA_SYNC_UPPER_THRESH) )
	  gbl_line_idx = sync_pos;
	else
	  gbl_line_idx = 0;
  }

  return(0);

} /* End of Detect_NOAA_Sync() */

/*------------------------------------------------------------------------*/

/*  Detect_Meteor_Sync()
 *
 *  Detects the sync train of a Meteor type APT signal.
 *  This is 5 cycles at 1200 Hz on the 2.4 KHz sub-carrier.
 */

  int
Detect_Meteor_Sync( int search_limit )
{
  /* FIFO Buffer for sync detection */
  static int sync_fifo[METEOR_SYNC_FIFO_SIZE];

  short
	sample_a,      /* First of the pair of carrier samples  */
	sample_b;      /* Second of the pair of carrier samples */

  int
	sync_idx,      /* Counter used in detecting sync train position  */
	carrier_val,   /* Instantaneous value of 2.4 Khz sub-carrier     */
	sync_val,      /* Value produced by sync detection algorithm     */
	sync_max,      /* Max value produced by sync detection algorithm */
	sync_ave,      /* Average value of carrier samples in sync fifo  */
	sync_pos,      /* Position in sample buffer where sync o/p peaks */
	sync_cnt,      /* Counter of sync pulses processed */
	peak_idx,      /* Index used in searching for peak in sync value */
	fifo_idx,      /* Index for sync detection fifo */
	idx;           /* Index for various loops etc.  */


  /* Move through buffer until sync detector  */
  /* peaks. On exit, the buffer index should  */
  /* be at the end of the sync pulse sequence */

  sync_max = fifo_idx = 0;
  sync_pos = gbl_line_idx;
  search_limit--;

  for( peak_idx = 0; peak_idx < search_limit; peak_idx++ )
  {
	/* Read first of the pair of carrier samples */
	sample_a = gbl_line_buffer[gbl_line_idx++];

	/* Read second of the pair of carrier samples */
	sample_b = gbl_line_buffer[gbl_line_idx];

	/* This is the instantaneous magnitude of the 2.4 Khz  */
	/* subcarrier, calculated as the scalar magnitude of   */
	/* two consecutive audio samples at 90 deg phase angle */
	carrier_val = Carrier_Amplitude( sample_a, sample_b );

	/* Store samples in fifo in cyclic manner */
	if( ++fifo_idx == METEOR_SYNC_FIFO_SIZE )
	  fifo_idx = 0;
	sync_fifo[fifo_idx] = carrier_val;

	/* Evaluate the average of the samples in fifo, to */
	/* be used as a zero reference in summation below. */
	sync_val = 0;
	for( idx = 0; idx < METEOR_SYNC_FIFO_SIZE; idx++ )
	  sync_val += sync_fifo[idx];
	sync_ave = sync_val/METEOR_SYNC_FIFO_SIZE;

	/* Subtract 5x2-sample groups and add 5x4-sample groups */
	/* to the sync accumulator alternately, simulating      */
	/* synchromous detection. The sync train appears to be  */
	/* 5 cycles at 1200 Hz with a 2/1 mark/space ratio.     */
	/* Their average value is used as a zero reference.     */
	sync_val = 0;
	sync_idx = fifo_idx;
	for( sync_cnt = 0; sync_cnt < METEOR_SYNC_PULSES; sync_cnt++ )
	{
	  /* Subtract carrier samples from accumulator */
	  for( idx = 0; idx < METEOR_SYNC_MARK; idx++ )
	  {
		sync_val -= sync_fifo[fifo_idx] - sync_ave;
		if( ++sync_idx >= METEOR_SYNC_FIFO_SIZE )
		sync_idx = 0;
	  }

	  /* Add carrier samples into accumulator */
	  for( idx = 0; idx < METEOR_SYNC_SPACE; idx++ )
	  {
		sync_val += sync_fifo[fifo_idx] - sync_ave;
		if( ++sync_idx >= METEOR_SYNC_FIFO_SIZE )
		  sync_idx = 0;
	  }

	} /* for( sync_cnt = 0; sync_cnt < METEOR_SYNC_PULSES; sync_cnt++ ) */

	/* Record position in buffer where sync detector peaks */
	if( sync_val > sync_max)
	{
	  sync_max = sync_val;
	  sync_pos = gbl_line_idx;
	}

  } /* End of for( peak_idx = 0; peak_idx < search_limit-1; peak_idx++ ) */

  /* Decoding from DSP */
  if( isFlagClear(ACTION_PROCESS_FILE) )
  {
	/* Adjust sync detector progress bar */
	gdouble fr = (gdouble)sync_max/(gdouble)METEOR_SYNC_UPPER_THRESH;
	if( fr > 1.0 ) fr = 1.0;
	gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR(gbl_sync_pbar), fr );

	/* Initialize skew label */
	gtk_label_set_text( GTK_LABEL(lookup_widget(
			xwxapt_main_window, "err_label")), "?" );

	/* Setup buffer index to max sync position */
	/* if sync maximum is within the threshold */
	if( (sync_max > METEOR_SYNC_LOWER_THRESH) &&
		(sync_max < METEOR_SYNC_UPPER_THRESH) )
	{
	  char buff[12];

	  /* Display sync error */
	  gbl_line_idx = sync_pos;
	  gbl_sync_err = sync_pos - gbl_sync_ref;
	  snprintf( buff, 12, "%d", gbl_sync_err );
	  gtk_label_set_text( GTK_LABEL(lookup_widget(
			  xwxapt_main_window, "err_label")), buff );

	  /* If sync position is in range, show 'Lock' */
	  if( (abs( gbl_sync_err ) < SYNC_DETECT) &&
		  isFlagClear(ACTION_SYNC_BUFFER) )
	  {
		if( isFlagClear(ICON_SYNC_LOCK) )
		{
		  ClearFlag( ICON_FLAGS_ALL );
		  SetFlag( ICON_SYNC_LOCK );
		  Set_Sync_Icon( "gtk-yes" );
		}
	  }
	  else /* Show 'Search' */
	  {
		if( isFlagClear(ICON_SYNC_SEARCH) )
		{
		  ClearFlag( ICON_FLAGS_ALL );
		  SetFlag( ICON_SYNC_SEARCH );
		  Set_Sync_Icon( "gtk-dialog-question" );
		}

	  } /* if( abs(gbl_line_idx - gbl_sync_ref) < 3 ) */
	}
	else /* If sync level is out of range, show 'Unlock' */
	{
	  gbl_line_idx = 0;

	  if( isFlagClear(ICON_SYNC_UNLOCK) )
	  {
		ClearFlag( ICON_FLAGS_ALL );
		SetFlag( ICON_SYNC_UNLOCK );
		Set_Sync_Icon( "gtk-no" );
	  }

	} /* if( (sync_max > METEOR_SYNC_LOWER_THRESH) && */

  } /* if( isFlagClear(ACTION_PROCESS_FILE) ) */
  else
  {
	/* Setup buffer index to max sync position */
	/* if sync maximum is within the threshold */
	if( (sync_max > METEOR_SYNC_LOWER_THRESH) &&
		(sync_max < METEOR_SYNC_UPPER_THRESH) )
	  gbl_line_idx = sync_pos;
	else
	  gbl_line_idx = 0;
  }

  return(0);

} /* End of Detect_Meteor_Sync() */

/*------------------------------------------------------------------------*/

/*  Detect_Sync_Train()
 *
 *  Detects sync train position in line buffer
 *  depending on the type of satellite processed
 */

  int
Detect_Sync_Train( int search_limit )
{
  gbl_line_idx = 0;
  switch( gbl_sat_type )
  {
	case NOAA:
	  Detect_NOAA_Sync(search_limit);
	  break;
	case METEOR:
	  Detect_Meteor_Sync(search_limit);
  }

  return(gbl_line_idx);

} /* End of Detect_Sync_Train() */

/*------------------------------------------------------------------------*/

/* Carrier_Amplitude()
 *
 * The instantaneous amplitude a of the 2.4 KHz sub-carrier
 * is calculated from two consecutive DSP samples a1, a2
 * using a = sqrt(a1^2 + a2^2 -2*a1*a2*cosf)/sinf where
 * f = 2*pi*Freq_carrier/Freq_dsp = phase between samples
 */

 int
Carrier_Amplitude( short x, short y )
{
  int carrier_amplitude;	/* Amplitude of 2.4 KHz sub-carrier */
  double a1, a2;			/* Amplitude of two consecutive samples */
  static double cosf, sinf; /* cos and sin of phase between above   */
  static int init = TRUE;

  if( init )
  {
	/* Phase angle between two consecutive sub-carrer samples */
	double f = 2.0 * M_PI * (double)CARRIER_FREQ / (double)DSP_SPEED;
	cosf = cos(f);
	sinf = sin(f);
	init = FALSE;
  }

  /* Carrier amplitude */
  a1 = (double)x; a2 = (double)y;
  carrier_amplitude = (int)
	( sqrt((a1*a1 + a2*a2 - 2.0*a1*a2*cosf)/sinf) + 0.5 );

  return( carrier_amplitude );

} /* End of Carrier_Amplitude() */

/*------------------------------------------------------------------------*/

/*  Set_Sync_Icon()
 *
 * Changes the sync status icon
 */

  void
Set_Sync_Icon( gchar *icon_name )
{
  /* Remove previous icon */
  gtk_container_remove( GTK_CONTAINER(gbl_status_table), gbl_sync_icon );

  /* Attach a new icon */
  gbl_sync_icon = gtk_image_new_from_stock( icon_name, GTK_ICON_SIZE_BUTTON );
  gtk_widget_show( gbl_sync_icon );

  gtk_container_add( GTK_CONTAINER(gbl_status_table), gbl_sync_icon );
  gtk_misc_set_padding (GTK_MISC (gbl_sync_icon), 0, 4);

} /* Set_Sync_Icon( gchar *img_name ) */

/*------------------------------------------------------------------------*/
