/*  sound.c
 *
 *  Sound-card handling functions of xwxapt application
 */

/*
 *  xwxapt: An application to decode APT signals from weather
 *  satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include <gtk/gtk.h>
#include "xwxapt.h"
#include "interface.h"
#include "support.h"

/* Audio level scale  */
extern GtkWidget *gbl_level_pbar;

/* Satellite type names */
extern char *gbl_sat_names[];

/* List of required sync position references */
extern int gbl_sync_refs[];

extern int
  gbl_sat_type,   /* Weather satellite type number   */
  gbl_duration,   /* Time limit in sec on processing */
  gbl_line_idx,   /* Index to carrier samples buffer */
  gbl_sync_err, /* Deviation from sync reference   */
  gbl_sync_ref;   /* Required position of sync train */

int
  gbl_dsp_fd,     /* File descriptor of DSP device   */
  gbl_mix_fd;     /* File descriptor of mixer device */

/* Pointer to samples file */
extern FILE *gbl_samples_fp;

/* Carrier samples buffer for one line */
extern short gbl_line_buffer[];

/* Buffer for reading in DSP data */
unsigned char *gbl_dsp_buffer = NULL;

static int
  dsp_buf_idx,	/* Index for this buffer  */
  chunk_size,	/* DSP block (chunk) size */
  dsp_rate_err;	/* dsp sampling rate error*/

char
  gbl_samples_file[MAX_FILE_NAME+1],  /* Sample file name */
  gbl_image_file  [MAX_FILE_NAME+1];  /* Image file name  */

/* Runtime config data */
extern rc_data_t rc_data;

/*------------------------------------------------------------------------*/

/*  Setup_Sound_Card()
 *
 *  Sets up mixer and DSP device and adjusts input
 *  level for correct sampling of APT signal
 */

  void
Setup_Sound_Card( void )
{
  int
	rec_src,     /* Recording/Capture source (Line?)  */
	reclev_dev,  /* Recording Level device (IGAIN?)   */
	record_lev,  /* Recording level request to mixer  */
	dsp_speed,   /* DSP sampling speed (9600 Hz)      */
	sample_fmt,  /* DSP sample format, 8-bit unsigned */
	channels,    /* (Stereo) or Mono mode selection   */
	itmp;        /* Temporary int for tests etc       */

  /* Sound device names from souncard.h */
  char *dev_name[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_NAMES;

  /* Messages buffer */
  char message[68];

  /* File lock information */
  struct flock lockinfo;


  /* Return if sound card is ready */
  if( isFlagSet(SOUND_CARD_SETUP) )
	return;

  /*** Open and lock mixer and dsp devices ***/
  Show_Message( "Setting up the sound card ... ", "black" );

  /* Open mixer device, abort on error */
  snprintf( message, 68,
	  "Opening mixer device %s", rc_data.mix_dev );
  Show_Message( message, "black" );
  if( (gbl_mix_fd = open(rc_data.mix_dev, O_RDWR, 0)) == -1 )
  {
	snprintf( message, 68, "xwxapt: %s", rc_data.mix_dev );
	perror( message );
	snprintf( message, 68, "Failed to open %s", rc_data.mix_dev );
	Show_Message( message, "red" );
	Error_Dialog();
	return;
  }

  /* Attempt to lock entire mixer device file */
  snprintf( message, 68, "Locking mixer device %s", rc_data.mix_dev );
  Show_Message( message, "black" );
  lockinfo.l_type   = F_WRLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If mixer device is already locked, abort */
  if( fcntl( gbl_mix_fd, F_SETLK, &lockinfo ) < 0 )
  {
	snprintf( message, 68, "xwxapt: %s", rc_data.mix_dev );
	perror( message );
	fcntl( gbl_mix_fd, F_GETLK, &lockinfo );
	snprintf( message, 68,
		"Failed to lock %s\n"
		"Device already locked by pid %d",
		rc_data.mix_dev, lockinfo.l_pid);
	Show_Message( message, "red" );
	Error_Dialog();
	return;
  }

  /* Open dsp device, abort on error */
  snprintf( message, 68, "Opening dsp device %s", rc_data.dsp_dev );
  Show_Message( message, "black" );
  if( (gbl_dsp_fd = open(rc_data.dsp_dev, O_RDONLY, 0)) == -1 )
  {
	snprintf( message, 68, "xwxapt: %s", rc_data.dsp_dev );
	perror( message );
	snprintf( message, 68, "Failed to open %s", rc_data.dsp_dev );
	Show_Message( message, "red" );
	Error_Dialog();
	return;
  }

  /* Attempt to lock entire dsp device file */
  snprintf( message, 68, "Locking dsp device %s", rc_data.dsp_dev );
  Show_Message( message, "black" );
  lockinfo.l_type   = F_RDLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If dsp device is already locked, abort */
  if( fcntl( gbl_dsp_fd, F_SETLK, &lockinfo ) < 0 )
  {
	snprintf( message, 68, "xwxapt: %s", rc_data.dsp_dev );
	perror( message );
	fcntl( gbl_dsp_fd, F_GETLK, &lockinfo );
	snprintf( message, 68,
		"Failed to lock %s\n"
		"Device already locked by pid %d",
		rc_data.dsp_dev, lockinfo.l_pid);
	Show_Message( message, "red" );
	Error_Dialog();
	return;
  }

  /*** Set up sound card parameters ***/
  Show_Message( "Setting sound card parameters...", "black" );

  /* Find number of recording source matching name in xwxaptrc */
  for( rec_src = 0; rec_src < SOUND_MIXER_NRDEVICES; rec_src++ )
	if( strcmp( dev_name[rec_src], rc_data.rec_src ) == 0 )
	  break;
  if( rec_src == SOUND_MIXER_NRDEVICES )
  {
	snprintf( message, 68, "Failed to find device %s", rc_data.rec_src );
	Show_Message( message, "red" );
	Error_Dialog();
	return;
  }

  /* Make recording source mask */
  itmp = 1 << rec_src;

  /* Attempt to select recording source, abort on error */
  if( ioctl(gbl_mix_fd, SOUND_MIXER_WRITE_RECSRC, &itmp) == -1 )
  {
	perror("xwxapt: SOUND_MIXER_WRITE_RECSRC");
	snprintf( message, 68, "Failed to select device %s", rc_data.rec_src );
	Show_Message( message, "red" );
	Error_Dialog();
	return;
  }
  if( itmp != 1 << rec_src )
  {
	snprintf( message, 68, "Failed to select device %s", rc_data.rec_src );
	Show_Message( message, "red" );
	Error_Dialog();
	return;
  }

  /* Zero line input level to silence _my_ speakers. This
   * might cause other setups to silence input from APT receiver */
  record_lev = 0;
  if( ioctl(gbl_mix_fd, MIXER_WRITE(rec_src), &record_lev) == -1 )
  {
	perror("xwxapt: Setting Recording Source level");
	Show_Message( "Failed to set Recording Source level", "red" );
	Error_Dialog();
	return;
  }
  if( record_lev != 0 )
  {
	Show_Message( "Failed to set Recording Source level", "red" );
	Error_Dialog();
	return;
  }

  /* Find number of recording level device matching name in xwxaptrc */
  for( reclev_dev = 0; reclev_dev < SOUND_MIXER_NRDEVICES; reclev_dev++ )
	if( strcmp( dev_name[reclev_dev], rc_data.lev_dev ) == 0 )
	  break;
  if( reclev_dev ==  SOUND_MIXER_NRDEVICES )
  {
	snprintf( message, 68, "Failed to select device %s", rc_data.lev_dev );
	Show_Message( message, "red" );
	Error_Dialog();
	return;
  }

  /* Write default level to recording level device, abort on error */
  record_lev = itmp = rc_data.rec_lev | (rc_data.rec_lev << 8);
  if( ioctl(gbl_mix_fd, MIXER_WRITE(reclev_dev), &record_lev) == -1 )
  {
	perror("xwxapt: Setting Recording level");
	Show_Message( "Failed to set Recording level", "red" );
	Error_Dialog();
	return;
  }
  if( record_lev != itmp )
  {
	Show_Message( "Failed to set Recording level", "red" );
	Error_Dialog();
	return;
  }

  /* Set sample format to signed 16-bit, abort on error */
  sample_fmt = AFMT_S16_LE;
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_SETFMT, &sample_fmt) == -1 )
  {
	perror( "xwxapt: SNDCTL_DSP_SETFMT");
	Show_Message( "Failed to set dsp sample format", "red" );
	Error_Dialog();
	return;
  }
  if( sample_fmt != AFMT_S16_LE )
  {
	Show_Message( "Failed to set dsp sample format", "red" );
	Error_Dialog();
	return;
  }

  /* Set stereo or mono mode */
  channels = rc_data.num_chan;
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_CHANNELS, &channels) == -1 )
	perror("xwxapt: SNDCTL_DSP_CHANNELS");

  /* Setup DSP sampling speed, abort on error */
  dsp_speed = DSP_SPEED;
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_SPEED, &dsp_speed) == -1 )
  {
	perror( "xwxapt: SNDCTL_DSP_SPEED");
	Show_Message( "Failed to set dsp sampling speed", "red" );
	Error_Dialog();
	return;
  }
  if( dsp_speed != DSP_SPEED )
  {
	Show_Message( "Failed to set dsp sampling speed", "red" );
	Error_Dialog();
	return;
  }

  /* Read DSP block (chunck) size */
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_GETBLKSIZE, &chunk_size ) == -1 )
  {
	perror("xwxapt: Reading dsp block (chunk) size");
	Show_Message( "Failed to read dsp block (chunk) size", "red" );
	Error_Dialog();
	return;
  }

  /* Allocate memory to DSP data buffer */
  gbl_dsp_buffer = malloc( chunk_size );
  if( gbl_dsp_buffer == NULL )
  {
	Show_Message( "malloc() for samples buffer failed - Quit", "red" );
	Error_Dialog();
	return;
  }

  /* This will make Read_Dsp_Buffer() initialize */
  dsp_buf_idx = chunk_size;

  /* Succeeded, show a message and set flag */
  Show_Message( "Sound card set-up completed", "green");
  SetFlag( SOUND_CARD_SETUP );

  return;

} /* End of Setup_Sound_Card() */

/*------------------------------------------------------------------------*/

/*  Setup_Audio_Level()
 *
 *  Displays a simple graphical aid for setting
 *  up the audio level from the APT receiver
 */

  gboolean
Setup_Audio_Level( gpointer data )
{
  int
	carrier_max, /* Max value of carrier in  buffer */
	carrier_val, /* Sampled value of 2.4kHz carrier */
	idx;         /* Index used in loops etc */

  /* Progressbar fraction */
  double pbar_frac;

  /* Required max value of carrier level */
  static int level_max;

  /* First idle call */
  static gboolean first_call = TRUE;

  /* Setup sound card, abort on error */
  if( first_call )
  {
	Setup_Sound_Card();
	if( isFlagClear(SOUND_CARD_SETUP) )
	  return( FALSE );

	/* Select required max audio level */
	switch( gbl_sat_type )
	{
	  case NOAA:
		level_max = NOAA_CARRIER_MAX;
		break;

	  case METEOR:
		level_max = METEOR_CARRIER_MAX;
	}

	Show_Message( "Starting Audio Level Setup ...", "black" );
	first_call = FALSE;
  }

  /* Fill in the samples buffer from DSP, abort on error */
  if( Read_Dsp_Buffer(gbl_line_buffer, BLOCK_SIZE) == -1 )
  {
	Error_Dialog();
	first_call = TRUE;
	return( FALSE );
  }

  /*** Find max value of 2.4kHz subcarrier in buffer ***/
  carrier_max = 0;
  for( idx = 0; idx < BLOCK_SIZE-1; idx++ )
  {
	carrier_val = Carrier_Amplitude(
		gbl_line_buffer[idx],
		gbl_line_buffer[idx+1]);

	/* Find maximum carrier level */
	if( carrier_val > carrier_max )
	  carrier_max = carrier_val;
  }

  /* Show audio level on progress bar */
  pbar_frac = (gdouble)carrier_max / (gdouble)level_max ;
  if( pbar_frac > 1.0 )	pbar_frac = 1.0;
  gtk_progress_bar_set_fraction( GTK_PROGRESS_BAR(gbl_level_pbar), pbar_frac );

  if( isFlagSet(ACTION_SETUP_AUDIO) )
	return( TRUE );

  Show_Message( "Stopping Audio Level Setup", "black" );
  Cleanup();
  first_call = TRUE;

  return( FALSE );

} /* Setup_Audio_Level() */

/*------------------------------------------------------------------------*/

/*  Synchronize_Buffer()
 *
 *  Synchronizes samples buffer with the APT sync train
 */

  gboolean
Synchronize_Buffer( gpointer data )
{
  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Messages buffer */
  char message[42];

  /* Data block size used as read() argument  */
  int block_size;

  static int
	last_line_idx,		/* Saved value of samples buffer index */
	last_dsp_rate_err,	/* Saved value of DSP sampling rate error */
	rate_ok_cnt;		/* Count number of similar rate measurements */

  /* Cancel on user request */
  if( isFlagClear(ACTION_SYNC_BUFFER) )
  {
	Show_Message( "Search for APT sync train stopped", "black" );
	Cleanup();
	first_call = TRUE;
	return( FALSE );
  }

  /* Initialize on first call */
  if( first_call )
  {
	/* Setup sound card, abort on error */
	Setup_Sound_Card();
	if( isFlagClear(SOUND_CARD_SETUP) )
	  return( FALSE );

	/* Print some information */
	Show_Message( "Synchronizing with APT sync train ...", "black" );
	snprintf( message, 42, "Looking for a %s type sync train ... ",
		gbl_sat_names[gbl_sat_type] );
	Show_Message( message, "black" );

	/* Load appropriate sync reference */
	gbl_sync_ref = gbl_sync_refs[ gbl_sat_type ];

	dsp_rate_err = 0;
	last_line_idx = 0;
	last_dsp_rate_err = 0;
	rate_ok_cnt = 0;

	first_call = FALSE;
	SetFlag(ACTION_RATE_ERROR);

  } /* if( first_call ) */

  /*** Find DSP rate error ***/
  if( isFlagSet(ACTION_RATE_ERROR) )
  {
	/* Fill in the line buffer from DSP */
	if( Read_Dsp_Buffer(gbl_line_buffer, BLOCK_SIZE) == -1 )
	{
	  first_call = TRUE;
	  return( FALSE );
	}

	/* Locate sync train (sets gbl_line_idx just after it) */
	if( ! Detect_Sync_Train(BLOCK_SIZE) )
	  return( TRUE );

	/* DSP rate error is change in sync train position */
	dsp_rate_err = gbl_line_idx - last_line_idx;
	last_line_idx = gbl_line_idx;

	/* If measured dsp sampling rate error is
	 * within limits, signal end of measurement */
	if( abs(dsp_rate_err) > SYNC_ACCEPT )
	  rate_ok_cnt = 0;
	else
	  rate_ok_cnt++;
	if( rate_ok_cnt >= RATE_OK_COUNT )
	  ClearFlag(ACTION_RATE_ERROR);
	else
	  return( TRUE );

	/* Dummy-read samples to align buffer with sync ref. */
	block_size = gbl_line_idx - gbl_sync_ref;
	if( block_size < 0 )
	  block_size += BLOCK_SIZE;
	block_size += (dsp_rate_err * block_size)/BLOCK_SIZE;
	if( (Read_Dsp_Buffer(gbl_line_buffer, block_size) == -1) )
	{
	  Error_Dialog();
	  first_call = TRUE;
	  return( FALSE );
	}

	/* Show dsp rate error */
	snprintf( message, 42, "DSP sampling rate error: %d ppm",
		(1000000*dsp_rate_err)/BLOCK_SIZE );
	Show_Message( message, "orange" );
	gbl_sync_err = 0;

  } /* if( isFlagSet(ACTION_RATE_ERROR) ) */

  /*** Refine buffer alignment with sync reference ***/
  /* Dummy-read samples to align buffer with sync reference */
  block_size = BLOCK_SIZE + dsp_rate_err + gbl_sync_err;
  if( (Read_Dsp_Buffer(gbl_line_buffer, block_size) == -1) )
  {
	first_call = TRUE;
	return( FALSE );
  }

  /* Locate sync train. (sets gbl_line_idx just after it) */
  if( ! Detect_Sync_Train(BLOCK_SIZE) )
	return( TRUE );
  if( abs(gbl_sync_err) > SYNC_ACCEPT )
	return( TRUE );

  /* Print some information */
  Show_Message( "Synchronized with APT sync train", "green" );

  /* Start the selected process */
  if( isFlagSet(ACTION_PROCESS_DSP) )
	gtk_idle_add( Dsp_Process_Image, NULL );
  else
	if( isFlagSet(ACTION_RECORD_APT) )
	  gtk_idle_add( Record_APT_Signal, NULL );

  first_call = TRUE;
  ClearFlag(ACTION_SYNC_BUFFER);

  return( FALSE );

} /* End of Synchronize_Buffer() */

/*------------------------------------------------------------------------*/

/*  Record_APT_Signal()
 *
 *  Records samples of the 2.4 Khz subcarrier in a file
 */

  gboolean
Record_APT_Signal( gpointer data )
{
  /* Number of APT image lines processed */
  static int line_cnt;

  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Messages buffer */
  char message[300];

  /* Temp buffer to write line buffer to file */
  unsigned char buff[2];
  int idx;

  /* Samples written to file */
  ssize_t sample_cnt;


  if( first_call )
  {
	Show_Message( "Recording apt signal to file ...", "black" );

	/* Add extension to file name */
	strcat( gbl_samples_file, "-" );
	strcat( gbl_samples_file, gbl_sat_names[gbl_sat_type] );
	strcat( gbl_samples_file, ".bin" );

	/* Create a file for writing samples */
	gbl_samples_fp = fopen( gbl_samples_file, "w+" );
	if( gbl_samples_fp == NULL )
	{
	  perror( gbl_samples_file );
	  snprintf( message, 300,
		  "Failed to open recorded samples file:\n%s", gbl_samples_file );
	  Show_Message( message, "red" );
	  Error_Dialog();
	  return( FALSE );
	}

	/* Read samples from DSP, stop on user interrupt */
	/* or at time limit in number of APT lines read. */
	/* gbl_duration is in seconds of recording time. */
	snprintf( message, 300,
		"Setting recording duration to %d sec", gbl_duration );
	Show_Message( message, "black" );

	/* Print some information */
	snprintf( message, 300,
		"Recording APT signal in:\n%s ...", gbl_samples_file );
	Show_Message( message, "black" );

	line_cnt = 0;

	/* Write satellite type to samples file, abort on error */
	if( (sample_cnt = fwrite(&gbl_sat_type, 4, 1, gbl_samples_fp)) == -1 )
	{
	  perror( "xwxapt: fwrite()" );
	  snprintf( message, 300,
		  "Error writing to samples file\n%s", gbl_samples_file );
	  Show_Message( message, "red" );
	  Error_Dialog();
	  return( FALSE );
	}
	if( sample_cnt != 1 )
	{
	  snprintf( message, 300,
		  "Error writing to samples file\n%s", gbl_samples_file );
	  Show_Message( message, "red" );
	  Error_Dialog();
	  return( FALSE );
	}

	first_call = FALSE;

  } /* if( first_call ) */

  /* Fill samples buffer from DSP */
  Dsp_Fill_Buffer();

  /* Write carrier sample values to samples file, abort on error */
  for( idx = 0; idx < BLOCK_SIZE; idx++ )
  {
	buff[0] = gbl_line_buffer[idx] & 0xff;
	buff[1] = (gbl_line_buffer[idx] >> 8) & 0xff;

	if( (sample_cnt = fwrite(buff, 1, 2, gbl_samples_fp)) == -1 )
	{
	  perror( "xwxapt: fwrite()" );
	  snprintf(message, 300,
		  "Error writing to samples file\n%s", gbl_samples_file);
	  Show_Message( message, "red" );
	  Error_Dialog();
	  first_call = TRUE;
	  return( FALSE );
	}

	if( (sample_cnt != 2) )
	{
	  snprintf(message, 300,
		  "Error writing to samples file\n%s", gbl_samples_file);
	  Show_Message( message, "red" );
	  Error_Dialog();
	  first_call = TRUE;
	  return( FALSE );
	}
  } /* for( idx = 0; idx < BLOCK_SIZE; idx++ ) */

  /* Stop if line count reaches time */
  /* limit or on user stop request   */
  if( (++line_cnt/2 != gbl_duration) &&
	  isFlagSet(ACTION_RECORD_APT) )
	return( TRUE );

  /* Print some information */
  snprintf(message, 300,
	  "Closing samples file:\n%s", gbl_samples_file);
  Show_Message( message, "black" );

  Cleanup();
  first_call = TRUE;

  return( FALSE );

} /* End of Record_APT_Signal() */

/*------------------------------------------------------------------------*/

/*  File_Fill_Buffer()
 *
 *  Fills the audio sample buffer from source file
 */

  int
File_Fill_Buffer( void )
{
  /* Messages buffer */
  char message[300];

  /* For reading processed dsp data from file */
  ssize_t sample_cnt;

  /* Temp buffer to write line buffer to file */
  unsigned char buff[2];
  int idx;


  /* Fill in the samples buffer from file, return -1 on error */
  gbl_line_idx = 0;
  for( idx = 0; idx < BLOCK_SIZE; idx++ )
  {
	if( (sample_cnt = fread(buff, 1, 2, gbl_samples_fp)) == -1 )
	{
	  perror( "xwxapt: fread()" );
	  snprintf(message, 300,
		  "Error reading from samples file\n%s", gbl_samples_file);
	  Show_Message( message, "red" );
	  Error_Dialog();
	  return(-1);
	}

	if( feof(gbl_samples_fp) )
	  return(-1);
	else
	  if( sample_cnt != 2 )
	  {
		snprintf(message, 300,
			"Error reading from samples file\n%s", gbl_samples_file);
		Show_Message( message, "red" );
		Error_Dialog();
		return(-1);
	  }

	gbl_line_buffer[idx]  = buff[0];
	gbl_line_buffer[idx] |= buff[1] << 8;
	if( gbl_line_buffer[idx] & 0x8000 )
	  gbl_line_buffer[idx] -= 0x10000;

  } /* for( idx = 0; idx < BLOCK_SIZE; idx++ ) */

  /* Set buffer index just after sync train */
  /* Search for sync up to 1.5* sync refer. */
  if( ! Detect_Sync_Train(3*gbl_sync_refs[gbl_sat_type]/2) )
	return(0);

  /* If sync detection is far out, keep */
  /* buffer index in default position   */
  if( abs(gbl_line_idx - gbl_sync_ref) > SYNC_ACCEPT )
	gbl_line_idx = gbl_sync_ref;

  return(0);

} /* End of File_Fill_Buffer() */

/*------------------------------------------------------------------------*/

/*  Dsp_Fill_Buffer()
 *
 *  Fills the audio sample buffer from sound card DSP.
 *  Buffer filling is synchronized with APT sync train.
 */

  int
Dsp_Fill_Buffer( void )
{
  static int block_size = BLOCK_SIZE;

  /* Fill in the samples buffer from DSP */
  if( Read_Dsp_Buffer(gbl_line_buffer, block_size + dsp_rate_err ) == -1 )
  {
	Show_Message( "Error filling samples buffer from dsp", "red" );
	Error_Dialog();
	return(-1);
  }

  /* Search for sync up to 1.5 * sync ref, return if sync fails */
  if( ! Detect_Sync_Train(3*gbl_sync_ref/2) )
  {
	gbl_line_idx = gbl_sync_ref;
	return(0);
  }

  /* Try to keep APT sync train at beginning of buffer. This  */
  /* done done by increasing or decreasing the data  block    */
  /* size argument passed to read(), by adding or subtracting */
  /* half the difference between detected sync position and a */
  /* reference value. Error is limited to +-2 to avoid coarse */
  /* corrections to sync position due to noise interference.  */
  if( gbl_sync_err > SYNC_ACCEPT )
	gbl_sync_err = SYNC_ACCEPT;
  else
	if( gbl_sync_err < -SYNC_ACCEPT )
	  gbl_sync_err = -SYNC_ACCEPT;
  gbl_line_idx += gbl_sync_err;

  /* Vary the data block size argument to read() to
   * take up any drift in sync position due to errors
   * in the DSP sampling rate and satellite movement.
   * Only half the sync error is used to slow and
   * stabilize sync correction. */
  block_size = BLOCK_SIZE + gbl_sync_err/2;

  return(0);

} /* End of Dsp_Fill_Buffer() */

/*------------------------------------------------------------------------*/

/* Read_Dsp_Buffer()
 *
 * Reads the DSP's buffer and separates Left/Right
 * channel data depending on stereo or mono mode
 */

  int
Read_Dsp_Buffer( short *buffer, int buff_size )
{
  int
	sample_cnt,
	idx;

  static int stride;

  /* Transfer data from dsp buffer to buffer to be processed, */
  /* separating left or right channel data if in stereo mode  */
  if( buff_size > BUFFER_SIZE )
	buff_size -= BUFFER_SIZE;
  for( idx = 0; idx < buff_size; idx++ )
  {
	if( dsp_buf_idx >= chunk_size )
	{
	  /* Fill in the samples buffer from DSP, return -1 on error */
	  if( (sample_cnt = read(gbl_dsp_fd, gbl_dsp_buffer, chunk_size)) == -1 )
	  {
		perror( "xwxapt: read()" );
		Show_Message( "Error filling samples buffer from dsp", "red" );
		Error_Dialog();
		return(-1);
	  }

	  if( sample_cnt != chunk_size )
	  {
		perror( "xwxapt: read()" );
		Show_Message( "Error filling samples buffer from dsp", "red" );
		Error_Dialog();
		return(-1);
	  }

	  /* Start buffer index according to stereo|mono mode */
	  if( rc_data.num_chan == 1 ) /* Mono */
	  {
		dsp_buf_idx = 0;
		stride = 2;
	  }
	  else
	  {
		dsp_buf_idx = 2*rc_data.chan_used;
		stride = 4;
	  }

	} /* if( buff_index >= chunk_size ) */

	/* Fill buffer from left or right channel. dsp_buf_idx
	 * is arranged to point to the correct location */
	buffer[idx]  = gbl_dsp_buffer[dsp_buf_idx];
	buffer[idx] |= gbl_dsp_buffer[dsp_buf_idx+1] << 8;
	if( buffer[idx] & 0x8000 )
	  buffer[idx] -= 0x10000;

	/* Increment buffer index according to stereo|mono mode */
	dsp_buf_idx += stride;

  } /* for( idx = 0; idx < buff_size; idx++ ) */

  return(0);

} /* End of Read_Dsp_Buffer() */

/*------------------------------------------------------------------------*/

