/*  utils.c
 *
 *  Utility functions of xwxapt application
 */

/*
 *  xwxapt: An application to decode APT signals from
 *  weather satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "xwxapt.h"
#include "interface.h"
#include "support.h"
#include <gtk/gtk.h>

/* Text view scroller */
extern GtkWidget *gbl_text_scroller;

/* Text buffer for text view */
extern GtkTextBuffer *gbl_text_buffer;

extern int
  gbl_duration,   /* Time limit in sec on processing */
  gbl_dsp_fd,     /* File descriptor of DSP device   */
  gbl_mix_fd;     /* File descriptor of mixer device */

extern char
  gbl_samples_file[], /* Samples file name */
  gbl_image_file[];   /* Image files name  */

/* Buffer for reading in DSP data */
extern unsigned char *gbl_dsp_buffer;

/* Pointer to samples file */
extern FILE *gbl_samples_fp;

/* Runtime config data */
extern rc_data_t rc_data;


/*  Load_Config()
 *
 *  Loads the xwxaptrc configuration file
 */

  gboolean
Load_Config( gpointer data )
{
  char
	rc_fpath[64], /* File path to xwxaptrc */
	line[81];     /* Buffer for Load_Line  */

  /* Config file pointer */
  FILE *xwxaptrc;


  /* Setup file path to xwxaptrc and working dir */
  snprintf( rc_fpath, 64, "%s/xwxapt/xwxaptrc", getenv("HOME") );
  snprintf( rc_data.xwxapt_dir, 64, "%s/xwxapt/", getenv("HOME") );

  /* Open xwxaptrc file */
  xwxaptrc = fopen( rc_fpath, "r" );
  if( xwxaptrc == NULL )
  {
	perror( rc_fpath );
	Show_Message( "Failed to open xwxaptrc file", "red" );
	Error_Dialog();
	return( FALSE );
  }

  /*** Read runtime configuration data ***/
  /* Read mixer device, abort if EOF */
  if( Load_Line(line, xwxaptrc, "Mixer Device" ) != SUCCESS )
	return( FALSE );
  strncpy( rc_data.mix_dev, line, 14 );
  rc_data.mix_dev[14] = '\0';

  /* Read input DSP device, abort if EOF */
  if( Load_Line(line, xwxaptrc, "Input DSP Device") != SUCCESS )
	return( FALSE );
  strncpy( rc_data.dsp_dev, line, 14 );
  rc_data.dsp_dev[14] = '\0';

  /* Read recording source, abort if EOF */
  if( Load_Line(line, xwxaptrc, "Recording Source") != SUCCESS )
	return( FALSE );
  strncpy( rc_data.rec_src, line, 9 );
  rc_data.rec_src[9] = '\0';

  /* Read recording level device, abort if EOF */
  if( Load_Line(line, xwxaptrc, "Recording Level Device") != SUCCESS )
	return( FALSE );
  strncpy( rc_data.lev_dev, line, 9 );
  rc_data.lev_dev[9] = '\0';

  /* Read recording level, abort if EOF */
  if( Load_Line(line, xwxaptrc, "Recording Level") != SUCCESS )
	return( FALSE );
  rc_data.rec_lev = atoi( line );

  /* Read mode (S/M), abort if EOF */
  if( Load_Line(line, xwxaptrc, "STEREO|MONO Mode") != SUCCESS )
	return( FALSE );
  strncpy( rc_data.mode, line, 6 );
  rc_data.mode[6] = '\0';

  /* Read channel used, abort if EOF */
  if( Load_Line(line, xwxaptrc, "LEFT|RIGHT Channel") != SUCCESS )
	return( FALSE );
  strncpy( rc_data.channel, line, 5 );
  rc_data.channel[5] = '\0';

  /* Read default decode duration, abort if EOF */
  if( Load_Line(line, xwxaptrc, "Decoding Duration") != SUCCESS )
	return( FALSE );
  rc_data.def_dur = atoi( line );
  gbl_duration = rc_data.def_dur;
  /* Warn if decoding duration is too long */
  if( gbl_duration > 600 )
  {
	char message[134];
	snprintf( message, 134,
		"Default decoding duration specified\n"
		"in xwxaptrc (%d sec) seems excessive\n"
		"Default decoding duration limited to 600 sec",
		gbl_duration );
	Show_Message( message, "red" );
	gbl_duration = 600;
	Error_Dialog();
  }

  /*** Do the global config variables ***/
  /* Set stereo or mono mode */
  if( strcmp(rc_data.mode, "STEREO") == 0 )
	rc_data.num_chan = 2;
  else
	if( strcmp(rc_data.mode, "MONO") == 0 )
	  rc_data.num_chan = 1;
	else
	{
	  Show_Message( "No Stereo/Mono setting found in xwxaptrc", "red" );
	  Error_Dialog();
	  return( FALSE );
	}

  /* Channel in use left or right */
  if( strcmp(rc_data.channel, "RIGHT") == 0 )
	rc_data.chan_used = 1;
  else
	if( strcmp(rc_data.channel, "LEFT") == 0 )
	  rc_data.chan_used = 0;
	else
	{
	  Show_Message( "No Left/Right setting found in xwxaptrc", "red" );
	  Error_Dialog();
	  return( FALSE );
	}

  /* Load pseudo-colorization maps */
  if( ! Load_Colormap(xwxaptrc, &rc_data.noaa_A_map) )
  {
	  Show_Message( "Error loading colorization maps from xwxaptrc", "red" );
	  Error_Dialog();
	  return( FALSE );
  }

  if( ! Load_Colormap(xwxaptrc, &rc_data.meteor_map) )
  {
	  Show_Message( "Error loading colorization maps from xwxaptrc", "red" );
	  Error_Dialog();
	  return( FALSE );
  }

  /* Initialize top window etc */
  Initialize_Top_Window();

  fclose( xwxaptrc );

  return( FALSE );

} /* End of Load_Config() */

/*------------------------------------------------------------------*/

/* Load_Colormap()
 *
 * Loads the pseudo-colorization maps from rc file
 */

  gboolean
Load_Colormap( FILE *xwxaptrc, color_map_t *map )
{
  /* Buffer for Load_Line  */
  char line[81];
  int idx, match;

  /* Read number of ranges */
  if( Load_Line(line, xwxaptrc, "Colormap Number of Ranges") != SUCCESS )
	return( FALSE );
  map->num_ranges = atoi(line);

  /* Read color map */
  for( idx = 0; idx < map->num_ranges; idx++ )
  {
	if( Load_Line(line, xwxaptrc, "Color Map") != SUCCESS )
	  return( FALSE );

	match = sscanf( line,
		"%hhu %hhu %hhu %hhu %hhu %hhu %hhu %hhu",
		&map->gray_from[idx],  &map->gray_to[idx],
		&map->red_from[idx],   &map->red_to[idx],
		&map->green_from[idx], &map->green_to[idx],
		&map->blue_from[idx],  &map->blue_to[idx] );

	if( match != 8 )
	{
	  perror( "xwxapt" );
	  return( FALSE );
	}
  } /* for( idx = 0; idx < map->num_ranges; idx++ ) */

  /* Read white (cloud) threshold */
  if( Load_Line(line, xwxaptrc, "White (cloud) Threshold") != SUCCESS )
	return( FALSE );
  map->white_thld = atoi(line);

  return( TRUE );
}

/*------------------------------------------------------------------*/

/*  Load_Line()
 *
 *  Loads a line from a file, aborts on failure. Lines beginning
 *  with a '#' are ignored as comments. At the end of file EOF is
 *  returned. Lines assumed maximum 80 characters long.
 */

  int
Load_Line( char *buff, FILE *pfile, char *messg )
{
  int
	num_chr, /* Number of characters read, excluding lf/cr */
	chr;     /* Character read by getc() */
  char err_msg[128];

  /* Prepare error message */
  snprintf( err_msg, 128,
	  "Error reading %s\n"
	  "Premature EOF (End Of File)", messg );

  /* Clear buffer at start */
  buff[0] = '\0';
  num_chr = 0;

  /* Get next character, return error if chr = EOF */
  if( (chr = fgetc(pfile)) == EOF )
  {
	fprintf( stderr, "xwxapt: %s\n", err_msg );
	Show_Message( err_msg, "red" );
	Error_Dialog();
	return( EOF );
  }

  /* Ignore commented lines and eol/cr and tab */
  while(
	  (chr == '#') ||
	  (chr == HT ) ||
	  (chr == CR ) ||
	  (chr == LF ) )
  {
	/* Go to the end of line (look for LF or CR) */
	while( (chr != CR) && (chr != LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xwxapt: %s\n", err_msg );
		Show_Message( err_msg, "red" );
		Error_Dialog();
		return( EOF );
	  }

	/* Dump any CR/LF remaining */
	while( (chr == CR) || (chr == LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xwxapt: %s\n", err_msg );
		Show_Message( err_msg, "red" );
		Error_Dialog();
		return( EOF );
	  }

  } /* End of while( (chr == '#') || ... */

  /* Continue reading characters from file till
   * number of characters = 80 or EOF or CR/LF */
  while( num_chr < 80 )
  {
	/* If LF/CR reached before filling buffer, return line */
	if( (chr == LF) || (chr == CR) ) break;

	/* Enter new character to line buffer */
	buff[num_chr++] = chr;

	/* Get next character */
	if( (chr = fgetc(pfile)) == EOF )
	{
	  /* Terminate buffer as a string if chr = EOF */
	  buff[num_chr] = '\0';
	  return( SUCCESS );
	}

	/* Abort if end of line not reached at 80 char. */
	if( (num_chr == 80) && (chr != LF) && (chr != CR) )
	{
	  /* Terminate buffer as a string */
	  buff[num_chr] = '\0';
	  snprintf( err_msg, 128,
		  "Error reading %s\n"
		  "Line longer than 80 characters", messg );
	  fprintf( stderr, "xwxapt: %s\n%s\n", err_msg, buff );
	  Show_Message( err_msg, "red" );
	  Error_Dialog();
	  return( ERROR );
	}

  } /* End of while( num_chr < max_chr ) */

  /* Terminate buffer as a string */
  buff[num_chr] = '\0';

  return( SUCCESS );

} /* End of Load_Line() */

/*------------------------------------------------------------------------*/

/*  File_Name()
 *
 *  Prepare a file name, use date and time if null argument
 */

  void
File_Name( char *file_name )
{
  int
	len, /* String length of file_name */
	idx; /* Index to while loop */

  /* If file_name is null, use date and time as file name */
  if( strlen(file_name) == 0 )
  {
	/* Variables for reading time (UTC) */
	time_t tp;
	struct tm utc;

	/* Prepare a file name as UTC date-time. */
	/* Default paths are images/ and record/ */
	time( &tp );
	utc = *gmtime( &tp );

	strcpy( file_name, rc_data.xwxapt_dir );
	len = strlen( file_name );

	if( isFlagSet(ACTION_PROCESS_DSP) )
	  strftime( &file_name[len], 24, "images/%d%b%Y-%H%M", &utc );
	else
	  if( isFlagSet(ACTION_RECORD_APT) )
		strftime( &file_name[len], 24, "record/%d%b%Y-%H%M", &utc );
  }
  else /* Remove leading spaces from file_name */
  {
	idx = 0;
	len = strlen(file_name);
	do idx++;
	while( (file_name[idx] == ' ') && (idx != len) );
	file_name += idx - 1;
  }

} /* End of File_Name() */

/*------------------------------------------------------------------------*/

/* Fname()
 *
 * Finds file name in a file path
 */

  char *
Fname( char *fpath )
{
  int idx;

  idx = strlen( fpath );

  while( (--idx >= 0) && (fpath[idx] != '/') );

  return( &fpath[++idx] );

} /* Fname() */

/*------------------------------------------------------------------------*/

/*  Usage()
 *
 *  Prints usage information
 */

  void
Usage( void )
{
  fprintf( stderr, "%s\n", "Usage: xwxapt [-hv]" );

  fprintf( stderr, "%s\n",
	  "       -h: Print this usage information and exit");

  fprintf( stderr, "%s\n",
	  "       -v: Print version number and exit");

} /* End of Usage() */

/*------------------------------------------------------------------------*/

/*  Show_Message()
 *
 *  Prints a message string in the Text View scroller
 */

  void
Show_Message( char *message, char *attr )
{
  GtkAdjustment *adjustment;

  static GtkTextIter iter;
  static gboolean first_call = TRUE;

  /* Initialize */
  if( first_call )
  {
	first_call = FALSE;
	gtk_text_buffer_get_iter_at_offset( gbl_text_buffer, &iter, 0 );
  }

  /* Print message */
  gtk_text_buffer_insert( gbl_text_buffer, &iter, "\n", -1 );
  gtk_text_buffer_insert_with_tags_by_name(
	  gbl_text_buffer, &iter,
	  message, -1, attr, NULL );

  /* Scroll Text View to bottom */
  adjustment = gtk_scrolled_window_get_vadjustment
	( GTK_SCROLLED_WINDOW(gbl_text_scroller) );
  gtk_adjustment_set_value( adjustment, adjustment->upper );

  /* Wait for GTK to complete its tasks */
  while( g_main_context_iteration(NULL, FALSE) );

} /* End of Show_Message() */

/*------------------------------------------------------------------------*/

/* Realloc()
 *
 * Re-allocates memory for buffers
 */

  gboolean
Realloc( unsigned char **buffer, size_t size )
{
  /* Re-allocate memory to image buffers, abort on error */
  *buffer = realloc( *buffer, size );
  if( *buffer == NULL )
  {
	perror( "xwxapt: A memory allocation request failed" );
	Show_Message( "A memory allocation request failed - Quit", "red" );
	Error_Dialog();
	return(FALSE );
  }

  return( TRUE );

} /* End of Realloc() */

/*------------------------------------------------------------------------*/

/* Open_File()
 *
 * Opens a file, aborts on error
 */

  gboolean
Open_File( FILE **fp, char *fname, const char *mode )
{
  /* Message buffer */
  char message[65];

  /* Open Channel A image file */
  *fp = fopen( fname, mode );
  if( *fp == NULL )
  {
	perror( fname );
	snprintf( message, 65, "Failed to open file\n%s", Fname(fname) );
	Show_Message( message, "red" );
	Error_Dialog();
	return( FALSE );
  }

  return( TRUE );

} /* End of Open_File() */

/*------------------------------------------------------------------------*/

/* File_Image()
 *
 * Write an image buffer to file
 */

gboolean
File_Image(
	FILE *fp, const char *type,
	int width, int height,
	int max_val, unsigned char *buffer)
{
  int size;

  /* Write header in Ch-A output PPM files */
  if( fprintf(fp, "%s\n%s\n%d %d\n%d\n",
		type, "# Created by xwxapt", width, height, max_val) < 0 )
  {
	perror( "xwxapt: Error writing image to file" );
	Show_Message( "Error writing image to file", "red" );
	Error_Dialog();
	return( FALSE );
  }

  /* P6 type (PPM) files are 3* size in pixels */
  if( strcmp(type, "P6") == 0 )
	size = 3 * width * height;
  else
	size = width * height;

  /* Write image buffer to file, abort on error */
  if( fwrite(buffer, 1, size, fp) != size )
  {
	perror( "xwxapt: Error writing image to file" );
	Show_Message( "Error writing image to file", "red" );
	Error_Dialog();
	return( FALSE );
  }

  return( TRUE );

} /* File_Image() */

/*------------------------------------------------------------------------*/

/*  Cleanup()
 *
 *  Cleanup before quitting or not using sound card
 */

  void
Cleanup( void )
{
  if( gbl_mix_fd > 0 )
  {
	close( gbl_mix_fd );
	gbl_mix_fd = 0;
  }

  if( gbl_dsp_fd > 0 )
  {
	close( gbl_dsp_fd );
	gbl_dsp_fd = 0;
  }

  if( gbl_samples_fp != NULL )
  {
	fclose( gbl_samples_fp );
	gbl_samples_fp = NULL;
  }

  if( gbl_dsp_buffer != NULL )
  {
	free( gbl_dsp_buffer );
	gbl_dsp_buffer = NULL;
  }

  gbl_image_file[0]   = '\0';
  gbl_samples_file[0] = '\0';

  /* Clear status indications */
  Set_Sync_Icon( "gtk-stop" );

  ClearFlag( SOUND_CARD_SETUP );
  ClearFlag( ACTION_FLAGS_ALL );
  ClearFlag( ICON_FLAGS_ALL );

} /*  Cleanup() */

/*------------------------------------------------------------------------*/


/* Functions for testing and setting/clearing flags */

/* An int variable holding the single-bit flags */
static int Flags = 0;

  int
isFlagSet(int flag)
{
  return (Flags & flag);
}

  int
isFlagClear(int flag)
{
  return (~Flags & flag);
}

  void
SetFlag(int flag)
{
  Flags |= flag;
}

  void
ClearFlag(int flag)
{
  Flags &= ~flag;
}

  void
ToggleFlag(int flag)
{
  Flags ^= flag;
}

/*------------------------------------------------------------------*/

