/*  xwxapt.h
 *
 *  Header file for xwxapt.
 */

/*
 *  xwxapt: An application to decode APT signals from weather
 *  satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


#ifndef XWXAPT_H
#define XWXAPT_H 1

#include <gtk/gtk.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <stdio.h>
#include <time.h>

/* Color map data */
typedef struct
{
  /* Pseudo-colorization maps */
  unsigned char
	gray_from[6],	gray_to[6],
	red_from[6],	red_to[6],
	green_from[6],	green_to[6],
	blue_from[6],	blue_to[6];

  int
	white_thld,  /* White (clouds) threshold */
	num_ranges;  /* Number of grayscale ranges */

} color_map_t;

/* Runtime config data */
typedef struct
{
  char
	xwxapt_dir[64], /* xwxapt working directory */
	mix_dev[15], 	/* Mixer device     */
	dsp_dev[15], 	/* Input DSP device */
	rec_src[10], 	/* Capture source   */
	lev_dev[10], 	/* Rec level device */
	mode[7],     	/* Mono | Stereo    */
	channel[6];  	/* Left | Right     */

  int
	def_dur,	/* Default duration of decoding */
	rec_lev,	/* Capture/record level	*/
	num_chan,	/* Stereo(=2) or Mono(=1) selection */
	chan_used,	/* Channel in use: left=0, right=1  */
	dsp_in_fd,	/* Capture dsp fd  */
	mix_fd;		/* Mixer device fd */

  /* Color map data */
  color_map_t
	noaa_A_map,
	meteor_map;

} rc_data_t;

/* Special characters */
#define LF      0x0A /* Line Feed */
#define CR      0x0D /* Carriage Return */
#define HT      0x09 /* Horizontal Tab  */

/* Return values */
#define ERROR		1
#define SUCCESS		0

/* Generel definitions for image processing */
#define CARRIER_FREQ	  2400 /* APT carrier frequency (2.4KHz)     */
#define DSP_SPEED        48000 /* DSP recording speed in samples/sec */
#define BUFFER_SIZE      24200 /* Audio buffer size=samples/line+200 */
#define BLOCK_SIZE       24000 /* Read() block size=samples/line     */
#define SYNC_DETECT         16 /* Sync-detected threshold (samples)  */
#define SYNC_ACCEPT          8 /* Accept sample rate error (carrier samples) */
#define RATE_OK_COUNT        8 /* Number of consecutive rate measurements    */
#define BLACK_CUT_OFF        1 /* Black cut-off percentile for normalization */
#define WHITE_CUT_OFF        2 /* White cut-off percentile for normalization */
#define MAX_FILE_NAME      255 /* Max length for optional filenames */

/* Definitions for NOAA image processing */
#define NOAA_PIXEL_SAMPLES		       14 /* DSP samples used for an image pixel*/
#define NOAA_IMAGE_WIDTH              748 /* Width in pixels of images produced */
#define NOAA_SYNC_CHA_SPACE           540 /* Space between ch A sync and image  */
#define NOAA_CHA_CHB_SPACE           1522 /* Space between ch A & B, in samples */
#define NOAA_SYNC_PULSES	 		    7 /* Number of pulses in the sync train */
#define NOAA_SYNC_SAMPLES	           23 /* Carrier samples per 1/2 sync cycle */
#define NOAA_SYNC_TAIL			       90 /* Sync train tail length, in samples */
#define NOAA_SYNC_FIFO_SIZE           412 /* Size of sync detector samples fifo */
#define NOAA_SYNC_LOWER_THRESH    2600000 /* Sync train detector lower threshold*/
#define NOAA_SYNC_UPPER_THRESH    3100000 /* Sync train detector upper threshold*/
#define NOAA_SYNC_REF                 750 /* Required position of sync train    */
#define NOAA_SCALE_FACTOR              28 /* Factor to keep pixel values < 255  */
#define NOAA_CARRIER_MAX            35000 /* Max 2.4kHz carrier-level required  */

/* Definitions for Meteor image processing */
#define METEOR_PIXEL_SAMPLES	       24 /* DSP samples used for an image pixel*/
#define METEOR_IMAGE_WIDTH            800 /* Width in pixels of images produced */
#define METEOR_SYNC_IMAGE_SPACE       648 /* Space between sync and image start */
#define METEOR_SYNC_PULSES		        5 /* Number of pulses in the sync train */
#define METEOR_SYNC_MARK		       13 /* Number of carrier samples in mark  */
#define METEOR_SYNC_SPACE		       27 /* Number of carrier samples in space */
#define METEOR_SYNC_FIFO_SIZE         200 /* Size of sync detector samples fifo */
#define METEOR_SYNC_LOWER_THRESH  2000000 /* Sync train detector lower threshold*/
#define METEOR_SYNC_UPPER_THRESH  2500000 /* Sync train detector upper threshold*/
#define METEOR_SYNC_REF               648 /* Required position of sync train    */
#define METEOR_SCALE_FACTOR            12 /* Factor to keep pixel values < 255  */
#define METEOR_CARRIER_MAX            160 /* Max 2.4kHz carrier-level required  */

/* Satellite type definitions */
#define NOAA                      0 /* NOAA type APT format   */
#define METEOR                    1 /* METEOR type APT format */
#define NUM_SATS                  2 /* Number of APT formats supported */
#define SAT_NAMES  "NOAA", "Meteor"
#define SYNC_REFS   NOAA_SYNC_REF, METEOR_SYNC_REF

/* Color maps to use */
#define NOAA_A_COLORMAP 0
#define NOAA_B_COLORMAP 1
#define METEOR_COLORMAP 2

/* Channel to be used in audio capture */
#define LEFT                      0 /* Left channel  */
#define RIGHT                     1 /* Right channel */

/* Flow control flags */
#define ICON_SYNC_LOCK          0x0001 /* Sync detector locked icon     */
#define ICON_SYNC_UNLOCK        0x0002 /* Sync detector unlocked icon   */
#define ICON_SYNC_SEARCH        0x0004 /* Sync detector searching icon  */
#define ICON_SYNC_STOPPED       0x0008 /* Sync detector searching icon  */
#define ICON_FLAGS_ALL          0x000f /* All icon flags (for clearing) */
#define ACTION_PROCESS_DSP      0x0010 /* Process images from dsp       */
#define ACTION_PROCESS_FILE     0x0020 /* Process images from file      */
#define ACTION_RECORD_APT       0x0040 /* Record apt signal to file     */
#define ACTION_SETUP_AUDIO      0x0080 /* Run audio level setup func.   */
#define ACTION_SYNC_BUFFER      0x0100 /* Run buffer synchronizing func */
#define ACTION_RATE_ERROR       0x0200 /* Measure dsp sampling rate err */
#define ACTION_FLAGS_ALL        0x03f0 /* All action flags (clearing)   */
#define IMAGE_NORMALIZE         0x0400 /* Histogram normalize wx image  */
#define IMAGE_COLORIZE          0x0800 /* Histogram normalize wx image  */
#define IMAGE_ROTATE            0x1000 /* Rotate wx image 180 degrees   */
#define SOUND_CARD_SETUP        0x2000 /* Sound card has been set up    */
#define START_STOP_TIMER        0x4000 /* Start-Stop timer is set       */

/*---------------------------------------------------------------------*/

/* Function prototypes produced by cproto */
/* callbacks.c */
gboolean Cancel_Timer(gpointer data);
void Error_Dialog(void);
void File_Selection(void);
void Popup_Menu(void);
/* image.c */
int NOAA_Image_Line(unsigned char *image_line, int *max_gray_val);
int Meteor_Image_Line(unsigned char *image_line, int *max_gray_val);
gboolean Dsp_Process_Image(gpointer data);
gboolean File_Process_Image(gpointer data);
void Normalize(unsigned char *image_buffer, int image_size);
void Rotate(unsigned char *image_buffer, int image_size);
void Colorize(unsigned char *image_buffer_a, unsigned char *image_buffer_b,
	int image_size,	unsigned char *colorized_buffer, int map);
void Halfsize_Image(unsigned char *buffer, int width,
	int height, int offset, gboolean color);
void Halfsize_Row(unsigned char *buffer, int width, int line_cnt, int offset);
/* main.c */
int main(int argc, char *argv[]);
void Initialize_Top_Window(void);
/* signal.c */
int Detect_NOAA_Sync(int search_limit);
int Detect_Meteor_Sync(int search_limit);
int Detect_Sync_Train(int search_limit);
int Carrier_Amplitude(short x, short y);
void Set_Sync_Icon( gchar *icon_name );
/* sound.c */
void Setup_Sound_Card(void);
gboolean Setup_Audio_Level(gpointer data);
gboolean Synchronize_Buffer(gpointer data);
gboolean Record_APT_Signal(gpointer data);
int File_Fill_Buffer(void);
int Dsp_Fill_Buffer(void);
int Read_Dsp_Buffer(short *buffer, int buff_size);
/* utils.c */
int Load_Line(char *buff, FILE *pfile, char *messg);
gboolean Load_Colormap(FILE *xwxaptrc, color_map_t *map);
gboolean Load_Config(gpointer data);
void File_Name(char *file_name);
char *Fname(char *fpath);
void Usage(void);
void Show_Message(char *message, char *attr);
void Cleanup(void);
int isFlagSet(int flag);
int isFlagClear(int flag);
gboolean Realloc(unsigned char **buffer, size_t size);
gboolean Open_File(FILE **fp, char *fname, const char *mode);
gboolean File_Image(FILE *fp, const char *type, int width, int height,
					int max_val, unsigned char *buffer);
void SetFlag(int flag);
void ClearFlag(int flag);
void ToggleFlag(int flag);

/*-------------------------------------------------------------------*/

#endif /* End of xwxapt.h */

